@@ -35,11 +35,12 @@ __author__ = "Mathieu Fenniak"
 __author_email__ = "biziqe@mathieu.fenniak.net"
 
 import re
-from utils import readNonWhitespace, RC4_encrypt
-from utils import b_, u_, chr_, ord_
-from utils import PdfStreamError
-import filters
-import utils
+from .utils import readNonWhitespace, RC4_encrypt
+from .utils import b_, u_, chr_, ord_
+from .utils import PdfStreamError
+import warnings
+from . import filters
+from . import utils
 import decimal
 import codecs
 #import debugging
@@ -55,7 +56,7 @@ def readObject(stream, pdf):
         return readStringFromStream(stream)
     elif tok == b_('/'):
         # name object
-        return NameObject.readFromStream(stream)
+        return NameObject.readFromStream(stream, pdf)
     elif tok == b_('['):
         # array object
         return ArrayObject.readFromStream(stream, pdf)
@@ -84,7 +85,7 @@ def readObject(stream, pdf):
             return NumberObject.readFromStream(stream)
         peek = stream.read(20)
         stream.seek(-len(peek), 1) # reset to start
-        if re.match(b_(r"(\d+)\s(\d+)\sR[^a-zA-Z]"), peek) != None:
+        if re.match(b_(r"(\d+)\s+(\d+)\s+R[^a-zA-Z]"), peek) != None:
             return IndirectObject.readFromStream(stream, pdf)
         else:
             return NumberObject.readFromStream(stream)
@@ -203,9 +204,11 @@ class IndirectObject(PdfObject):
                 # stream has truncated prematurely
                 raise PdfStreamError("Stream has ended unexpectedly")
             if tok.isspace():
+                if not generation:
+                    continue
                 break
             generation += tok
-        r = stream.read(1)
+        r = readNonWhitespace(stream)
         if r != b_("R"):
             raise utils.PdfReadError("Error reading indirect object reference at byte %s" % utils.hexStr(stream.tell()))
         return IndirectObject(int(idnum), int(generation), pdf)
@@ -217,7 +220,7 @@ class FloatObject(decimal.Decimal, PdfObject):
         try:
             return decimal.Decimal.__new__(cls, utils.str_(value), context)
         except:
-            return decimal.Decimal.__new__(cls, utils.str_(value))
+            return decimal.Decimal.__new__(cls, str(value))
     def __repr__(self):
         if self == self.to_integral():
             return str(self.quantize(decimal.Decimal(1)))
@@ -233,9 +236,10 @@ class FloatObject(decimal.Decimal, PdfObject):
 
 
 class NumberObject(int, PdfObject):
-    def __init__(self, value):
-        int.__init__(value)
-        
+    
+    def __new__(cls, value):
+        return int.__new__(cls, value)
+            
     def as_numeric(self):
         return int(b_(repr(self)))
 
@@ -338,8 +342,9 @@ def readStringFromStream(stream):
                 tok = b_(")")
             elif tok == b_("\\"):
                 tok = b_("\\")
-            elif tok == b_(" "):
-                tok = b_(" ")
+            elif tok in (b_(" "), b_("/"), b_("%"), b_("<"), b_(">"), b_("["), b_("]")):
+                # odd/unnessecary escape sequences we have encountered
+                tok = b_(tok)
             elif tok.isdigit():
                 # "The number ddd may consist of one, two, or three
                 # octal digits; high-order overflow shall be ignored.
@@ -445,17 +450,17 @@ class NameObject(str, PdfObject):
     delimiterCharacters = b_("("), b_(")"), b_("<"), b_(">"), b_("["), b_("]"), b_("{"), b_("}"), b_("/"), b_("%")
 
     def __init__(self, data):
-        str.__init__(data)
+        str.__init__(super(NameObject, self).__new__(str, data))
 
     def writeToStream(self, stream, encryption_key):
         stream.write(b_(self))
 
-    def readFromStream(stream):
+    def readFromStream(stream, pdf):
         debug = False
-        if debug: print stream.tell()
+        if debug: print((stream.tell()))
         name = stream.read(1)
         if name != b_("/"):
-            raise utils.PdfReadError, "name read error"
+            raise utils.PdfReadError("name read error")
         while True:
             tok = stream.read(1)
             if not tok:
@@ -465,8 +470,18 @@ class NameObject(str, PdfObject):
                 stream.seek(-1, 1)
                 break
             name += tok
-        if debug: print name
-        return NameObject(name.decode('utf-8'))
+        if debug: print(name)
+        try:
+            return NameObject(name.decode('utf-8'))
+        except UnicodeDecodeError as e:
+            # Name objects should represent irregular characters
+            # with a '#' followed by the symbol's hex number
+            if not pdf.strict:
+                warnings.warn("Illegal character in Name Object", utils.PdfReadWarning)
+                return NameObject(name)
+            else:
+                raise utils.PdfReadError("Illegal character in Name Object")
+        
     readFromStream = staticmethod(readFromStream)
 
 
@@ -490,7 +505,7 @@ class DictionaryObject(dict, PdfObject):
     def update(self, arr):
         # note, a ValueError halfway through copying values
         # will leave half the values in this dict.
-        for k, v in arr.iteritems():
+        for k, v in list(arr.items()):
             self.__setitem__(k, v)
 
     def raw_get(self, key):
@@ -526,7 +541,7 @@ class DictionaryObject(dict, PdfObject):
         if metadata == None:
             return None
         metadata = metadata.getObject()
-        import xmp
+        from . import xmp
         if not isinstance(metadata, xmp.XmpInformation):
             metadata = xmp.XmpInformation(metadata)
             self[NameObject("/Metadata")] = metadata
@@ -541,7 +556,7 @@ class DictionaryObject(dict, PdfObject):
 
     def writeToStream(self, stream, encryption_key):
         stream.write(b_("<<\n"))
-        for key, value in self.items():
+        for key, value in list(self.items()):
             key.writeToStream(stream, encryption_key)
             stream.write(b_(" "))
             value.writeToStream(stream, encryption_key)
@@ -552,8 +567,7 @@ class DictionaryObject(dict, PdfObject):
         debug = False
         tmp = stream.read(2)
         if tmp != b_("<<"):
-            raise utils.PdfReadError, \
-                ("Dictionary read error at byte %s: stream must begin with '<<'" % utils.hexStr(stream.tell()))
+            raise utils.PdfReadError("Dictionary read error at byte %s: stream must begin with '<<'" % utils.hexStr(stream.tell()))
         data = {}
         while True:
             tok = readNonWhitespace(stream)
@@ -563,7 +577,7 @@ class DictionaryObject(dict, PdfObject):
                 # stream has truncated prematurely
                 raise PdfStreamError("Stream has ended unexpectedly")
 
-            if debug: print "Tok:",tok
+            if debug: print(("Tok:", tok))
             if tok == b_(">"):
                 stream.read(1)
                 break
@@ -572,11 +586,16 @@ class DictionaryObject(dict, PdfObject):
             tok = readNonWhitespace(stream)
             stream.seek(-1, 1)
             value = readObject(stream, pdf)
-            if data.has_key(key):
+            if not data.get(key):
+                data[key] = value
+            elif pdf.strict:
                 # multiple definitions of key not permitted
-                raise utils.PdfReadError, ("Multiple definitions in dictionary at byte %s for key %s" \
+                raise utils.PdfReadError("Multiple definitions in dictionary at byte %s for key %s" \
                                            % (utils.hexStr(stream.tell()), key))
-            data[key] = value
+            else:
+                warnings.warn("Multiple definitions in dictionary at byte %s for key %s" \
+                                           % (utils.hexStr(stream.tell()), key), utils.PdfReadWarning)
+
         pos = stream.tell()
         s = readNonWhitespace(stream)
         if s == b_('s') and stream.read(5) == b_('tream'):
@@ -588,19 +607,19 @@ class DictionaryObject(dict, PdfObject):
             assert eol in (b_("\n"), b_("\r"))
             if eol == b_("\r"):
                 # read \n after
-                if stream.read(1)  != '\n':
+                if stream.read(1)  != b_('\n'):
                     stream.seek(-1, 1)
             # this is a stream object, not a dictionary
-            assert data.has_key("/Length")
+            assert "/Length" in data
             length = data["/Length"]
-            if debug: print data
+            if debug: print(data)
             if isinstance(length, IndirectObject):
                 t = stream.tell()
                 length = pdf.getObject(length)
                 stream.seek(t, 0)
             data["__streamdata__"] = stream.read(length)
-            if debug: print "here"
-            #if debug: print debugging.printAsHex(data["__streamdata__"])
+            if debug: print("here")
+            #if debug: print(binascii.hexlify(data["__streamdata__"]))
             e = readNonWhitespace(stream)
             ndstream = stream.read(8)
             if (e + ndstream) != b_("endstream"):
@@ -617,13 +636,12 @@ class DictionaryObject(dict, PdfObject):
                     # we found it by looking back one character further.
                     data["__streamdata__"] = data["__streamdata__"][:-1]
                 else:
-                   # if debug: print "E", e, ndstream, debugging.toHex(end)
+                    if debug: print(("E", e, ndstream, debugging.toHex(end)))
                     stream.seek(pos, 0)
-                    raise utils.PdfReadError, \
-                        ("Unable to find 'endstream' marker after stream at byte %s." % utils.hexStr(stream.tell()))
+                    raise utils.PdfReadError("Unable to find 'endstream' marker after stream at byte %s." % utils.hexStr(stream.tell()))
         else:
             stream.seek(pos, 0)
-        if data.has_key("__streamdata__"):
+        if "__streamdata__" in data:
             return StreamObject.initializeFromDictionary(data)
         else:
             retval = DictionaryObject()
@@ -636,7 +654,7 @@ class TreeObject(DictionaryObject):
         DictionaryObject.__init__(self)
         
     def hasChildren(self):
-        return self.has_key('/First')
+        return '/First' in self
     
     def __iter__(self):
         return self.children()
@@ -657,7 +675,7 @@ class TreeObject(DictionaryObject):
         child = pdf.getReference(childObj)
         assert isinstance(child, IndirectObject)
         
-        if not self.has_key('/First'):
+        if '/First' not in self:
             self[NameObject('/First')] = child
             self[NameObject('/Count')] = NumberObject(0)
             prev = None
@@ -680,10 +698,10 @@ class TreeObject(DictionaryObject):
     def removeChild(self, child):
         childObj = child.getObject()
         
-        if not childObj.has_key(NameObject('/Parent')):
-            raise ValueError, "Removed child does not appear to be a tree item"
+        if NameObject('/Parent') not in childObj:
+            raise ValueError("Removed child does not appear to be a tree item")
         elif childObj[NameObject('/Parent')] != self:
-            raise ValueError, "Removed child is not a member of this tree"
+            raise ValueError("Removed child is not a member of this tree")
         
         found = False
         prevRef = None
@@ -695,7 +713,7 @@ class TreeObject(DictionaryObject):
         while cur != None:
             if cur == childObj:
                 if prev == None:
-                    if cur.has_key(NameObject('/Next')):
+                    if NameObject('/Next') in cur:
                         # Removing first tree node
                         nextRef = cur[NameObject('/Next')]
                         next = nextRef.getObject()
@@ -708,10 +726,10 @@ class TreeObject(DictionaryObject):
                         assert self[NameObject('/Count')] == 1
                         del self[NameObject('/Count')]
                         del self[NameObject('/First')]
-                        if self.has_key(NameObject('/Last')):
+                        if NameObject('/Last') in self:
                             del self[NameObject('/Last')]
                 else:
-                    if cur.has_key(NameObject('/Next')):
+                    if NameObject('/Next') in cur:
                         # Removing middle tree node
                         nextRef = cur[NameObject('/Next')]
                         next = nextRef.getObject()
@@ -730,7 +748,7 @@ class TreeObject(DictionaryObject):
             
             prevRef = curRef
             prev = cur
-            if cur.has_key(NameObject('/Next')):
+            if NameObject('/Next') in cur:
                 curRef = cur[NameObject('/Next')]
                 cur = curRef.getObject()
             else:
@@ -738,28 +756,28 @@ class TreeObject(DictionaryObject):
                 cur = None
        
         if not found:
-            raise ValueError, "Removal couldn't find item in tree"
+            raise ValueError("Removal couldn't find item in tree")
        
         del childObj[NameObject('/Parent')]
-        if childObj.has_key(NameObject('/Next')):
+        if NameObject('/Next') in childObj:
             del childObj[NameObject('/Next')]
-        if childObj.has_key(NameObject('/Prev')):
+        if NameObject('/Prev') in childObj:
             del childObj[NameObject('/Prev')]
 
     def emptyTree(self):
         for child in self:
             childObj = child.getObject()
             del childObj[NameObject('/Parent')]
-            if childObj.has_key(NameObject('/Next')):
+            if NameObject('/Next') in childObj:
                 del childObj[NameObject('/Next')]
-            if childObj.has_key(NameObject('/Prev')):
+            if NameObject('/Prev') in childObj:
                 del childObj[NameObject('/Prev')]
 
-        if self.has_key(NameObject('/Count')):
+        if NameObject('/Count') in self:
             del self[NameObject('/Count')]
-        if self.has_key(NameObject('/First')):
+        if NameObject('/First') in self:
             del self[NameObject('/First')]
-        if self.has_key(NameObject('/Last')):
+        if NameObject('/Last') in self:
             del self[NameObject('/Last')]
 
 
@@ -780,7 +798,7 @@ class StreamObject(DictionaryObject):
         stream.write(b_("\nendstream"))
 
     def initializeFromDictionary(data):
-        if data.has_key("/Filter"):
+        if "/Filter" in data:
             retval = EncodedStreamObject()
         else:
             retval = DecodedStreamObject()
@@ -792,7 +810,7 @@ class StreamObject(DictionaryObject):
     initializeFromDictionary = staticmethod(initializeFromDictionary)
 
     def flateEncode(self):
-        if self.has_key("/Filter"):
+        if "/Filter" in self:
             f = self["/Filter"]
             if isinstance(f, ArrayObject):
                 f.insert(0, NameObject("/FlateDecode"))
@@ -828,15 +846,16 @@ class EncodedStreamObject(StreamObject):
         else:
             # create decoded object
             decoded = DecodedStreamObject()
+            
             decoded._data = filters.decodeStreamData(self)
-            for key, value in self.items():
+            for key, value in list(self.items()):
                 if not key in ("/Length", "/Filter", "/DecodeParms"):
                     decoded[key] = value
             self.decodedSelf = decoded
             return decoded._data
 
     def setData(self, data):
-        raise utils.PdfReadError, "Creating EncodedStreamObject is not currently supported"
+        raise utils.PdfReadError("Creating EncodedStreamObject is not currently supported")
 
 
 class RectangleObject(ArrayObject):
@@ -906,7 +925,7 @@ class RectangleObject(ArrayObject):
         return self.getUpperRight_x() - self.getLowerLeft_x()
 
     def getHeight(self):
-        return self.getUpperRight_y() - self.getLowerLeft_x()
+        return self.getUpperRight_y() - self.getLowerLeft_y()
 
     lowerLeft = property(getLowerLeft, setLowerLeft, None, None)
     lowerRight = property(getLowerRight, setLowerRight, None, None)
@@ -932,17 +951,17 @@ class Destination(TreeObject):
         elif typ == "/FitR":
             (self[NameObject("/Left")], self[NameObject("/Bottom")],
                 self[NameObject("/Right")], self[NameObject("/Top")]) = args
-        elif typ in ["/FitH", "FitBH"]:
+        elif typ in ["/FitH", "/FitBH"]:
             self[NameObject("/Top")], = args
-        elif typ in ["/FitV", "FitBV"]:
+        elif typ in ["/FitV", "/FitBV"]:
             self[NameObject("/Left")], = args
-        elif typ in ["/Fit", "FitB"]:
+        elif typ in ["/Fit", "/FitB"]:
             pass
         else:
             raise utils.PdfReadError("Unknown Destination Type: %r" % typ)
             
     def getDestArray(self):
-        return ArrayObject([self.raw_get('/Page'), self['/Type']] + [self[x] for x in ['/Left','/Bottom','/Right','/Top','/Zoom'] if self.has_key(x)])
+        return ArrayObject([self.raw_get('/Page'), self['/Type']] + [self[x] for x in ['/Left', '/Bottom', '/Right', '/Top', '/Zoom'] if x in self])
         
     def writeToStream(self, stream, encryption_key):
         stream.write(b_("<<\n"))
@@ -1005,7 +1024,7 @@ class Destination(TreeObject):
 class Bookmark(Destination):
     def writeToStream(self, stream, encryption_key):
         stream.write(b_("<<\n"))
-        for key in [NameObject(x) for x in ['/Title', '/Parent', '/First', '/Last', '/Next', '/Prev'] if self.has_key(x)]:
+        for key in [NameObject(x) for x in ['/Title', '/Parent', '/First', '/Last', '/Next', '/Prev'] if x in self]:
             key.writeToStream(stream, encryption_key)
             stream.write(b_(" "))
             value = self.raw_get(key)
@@ -1078,7 +1097,7 @@ _pdfDocEncoding = (
 assert len(_pdfDocEncoding) == 256
 
 _pdfDocEncoding_rev = {}
-for i in xrange(256):
+for i in range(256):
     char = _pdfDocEncoding[i]
     if char == u_("\u0000"):
         continue
