@@ -27,9 +27,18 @@
 # ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 # POSSIBILITY OF SUCH DAMAGE.
 
-from generic import *
-from pdf import PdfFileReader, PdfFileWriter
-from StringIO import StringIO
+from .generic import *
+from .utils import string_type
+from .pdf import PdfFileReader, PdfFileWriter
+from .pagerange import PageRange
+from sys import version_info
+if version_info < ( 3, 0 ):
+    from cStringIO import StringIO
+    StreamIO = StringIO
+else:
+    from io import BytesIO
+    from io import FileIO as file
+    StreamIO = BytesIO
 
 class _MergedPage(object):
     """
@@ -77,8 +86,9 @@ class PdfFileMerger(object):
         You may prevent the source document's bookmarks from being imported by
         specifying "import_bookmarks" as False.
         
-        You may also use the "pages" parameter to merge only the specified range of 
-        pages from the source document into the output document.
+        The optional "pages" parameter can be a PageRange or a 
+        (start, stop[, step]) tuple to merge only the specified range of pages
+        from the source document into the output document.
         """
         
         # This parameter is passed to self.inputs.append and means
@@ -87,35 +97,37 @@ class PdfFileMerger(object):
         
         # If the fileobj parameter is a string, assume it is a path
         # and create a file object at that location. If it is a file,
-        # copy the file's contents into a StringIO stream object; if 
+        # copy the file's contents into a BytesIO (or StreamIO) stream object; if 
         # it is a PdfFileReader, copy that reader's stream into a 
-        # StringIO stream.
+        # BytesIO (or StreamIO) stream.
         # If fileobj is none of the above types, it is not modified
-        if type(fileobj) in (str, unicode):
+        if type(fileobj) == string_type:
             fileobj = file(fileobj, 'rb')
             my_file = True
-        elif type(fileobj) == file:
+        elif isinstance(fileobj, file):
             fileobj.seek(0)
             filecontent = fileobj.read()
-            fileobj = StringIO(filecontent)
+            fileobj = StreamIO(filecontent)
             my_file = True
-        elif type(fileobj) == PdfFileReader:
+        elif isinstance(fileobj, PdfFileReader):
             orig_tell = fileobj.stream.tell()   
             fileobj.stream.seek(0)
-            filecontent = StringIO(fileobj.stream.read())
+            filecontent = StreamIO(fileobj.stream.read())
             fileobj.stream.seek(orig_tell) # reset the stream to its original location
             fileobj = filecontent
             my_file = True
             
         # Create a new PdfFileReader instance using the stream
-        # (either file or StringIO) created above
+        # (either file or BytesIO or StringIO) created above
         pdfr = PdfFileReader(fileobj, strict=self.strict)
         
-        # Find the range of pages to merge
+        # Find the range of pages to merge.
         if pages == None:
             pages = (0, pdfr.getNumPages())
-        elif type(pages) in (int, float, str, unicode):
-            raise TypeError('"pages" must be a tuple of (start, end)')
+        elif isinstance(pages, PageRange):
+            pages = pages.indices(pdfr.getNumPages())
+        elif not isinstance(pages, tuple):
+            raise TypeError('"pages" must be a tuple of (start, stop[, step])')
         
         srcpages = []
         if bookmark:
@@ -176,7 +188,7 @@ class PdfFileMerger(object):
         kind of file-like object)
         """
         my_file = False
-        if type(fileobj) in (str, unicode):
+        if type(fileobj) in (str, str):
             fileobj = file(fileobj, 'wb')
             my_file = True
 
@@ -229,11 +241,12 @@ class PdfFileMerger(object):
 
     def _trim_dests(self, pdf, dests, pages):
         """
-        Removes any named destinations that are not a part of the specified page set
+        Removes any named destinations that are not a part of the specified 
+        page set.
         """
         new_dests = []
         prev_header_added = True
-        for k, o in dests.items():
+        for k, o in list(dests.items()):
             for j in range(*pages):
                 if pdf.getPage(j).getObject() == o['/Page'].getObject():
                     o[NameObject('/Page')] = o['/Page'].getObject()
@@ -244,12 +257,13 @@ class PdfFileMerger(object):
     
     def _trim_outline(self, pdf, outline, pages):
         """
-        Removes any outline/bookmark entries that are not a part of the specified page set
+        Removes any outline/bookmark entries that are not a part of the 
+        specified page set.
         """
         new_outline = []
         prev_header_added = True
         for i, o in enumerate(outline):
-            if type(o) == list:
+            if isinstance(o, list):
                 sub = self._trim_outline(pdf, o, pages)
                 if sub:
                     if not prev_header_added:
@@ -271,7 +285,7 @@ class PdfFileMerger(object):
         for v in dests:
             pageno = None
             pdf = None
-            if v.has_key('/Page'):
+            if '/Page' in v:
                 for i, p in enumerate(self.pages):
                     if p.id == v['/Page']:
                         v[NameObject('/Page')] = p.out_pagedata
@@ -289,13 +303,13 @@ class PdfFileMerger(object):
 
         last_added = None
         for b in bookmarks:
-            if type(b) == list:
+            if isinstance(b, list):
                 self._write_bookmarks(b, last_added)
                 continue
                 
             pageno = None
             pdf = None
-            if b.has_key('/Page'):
+            if '/Page' in b:
                 for i, p in enumerate(self.pages):
                     if p.id == b['/Page']:
                         #b[NameObject('/Page')] = p.out_pagedata
@@ -303,45 +317,45 @@ class PdfFileMerger(object):
                         #nothing more to add
                         #if b['/Type'] == '/Fit' or b['/Type'] == '/FitB'
                         if b['/Type'] == '/FitH' or b['/Type'] == '/FitBH':
-                            if b.has_key('/Top') and not isinstance(b['/Top'], NullObject):
+                            if '/Top' in b and not isinstance(b['/Top'], NullObject):
                                 args.append(FloatObject(b['/Top']))
                             else:
                                 args.append(FloatObject(0))
                             del b['/Top']
                         elif b['/Type'] == '/FitV' or b['/Type'] == '/FitBV':
-                            if b.has_key('/Left') and not isinstance(b['/Left'], NullObject):
+                            if '/Left' in b and not isinstance(b['/Left'], NullObject):
                                 args.append(FloatObject(b['/Left']))
                             else:
                                 args.append(FloatObject(0))
                             del b['/Left']
                         elif b['/Type'] == '/XYZ':
-                            if b.has_key('/Left') and not isinstance(b['/Left'], NullObject):
+                            if '/Left' in b and not isinstance(b['/Left'], NullObject):
                                 args.append(FloatObject(b['/Left']))
                             else:
                                 args.append(FloatObject(0))
-                            if b.has_key('/Top') and not isinstance(b['/Top'], NullObject):
+                            if '/Top' in b and not isinstance(b['/Top'], NullObject):
                                 args.append(FloatObject(b['/Top']))
                             else:
                                 args.append(FloatObject(0))
-                            if b.has_key('/Zoom') and not isinstance(b['/Zoom'], NullObject):
+                            if '/Zoom' in b and not isinstance(b['/Zoom'], NullObject):
                                 args.append(FloatObject(b['/Zoom']))
                             else:
                                 args.append(FloatObject(0))
                             del b['/Top'], b['/Zoom'], b['/Left']
                         elif b['/Type'] == '/FitR':
-                            if b.has_key('/Left') and not isinstance(b['/Left'], NullObject):
+                            if '/Left' in b and not isinstance(b['/Left'], NullObject):
                                 args.append(FloatObject(b['/Left']))
                             else:
                                 args.append(FloatObject(0))
-                            if b.has_key('/Bottom') and not isinstance(b['/Bottom'], NullObject):
+                            if '/Bottom' in b and not isinstance(b['/Bottom'], NullObject):
                                 args.append(FloatObject(b['/Bottom']))
                             else:
                                 args.append(FloatObject(0))
-                            if b.has_key('/Right') and not isinstance(b['/Right'], NullObject):
+                            if '/Right' in b and not isinstance(b['/Right'], NullObject):
                                 args.append(FloatObject(b['/Right']))
                             else:
                                 args.append(FloatObject(0))
-                            if b.has_key('/Top') and not isinstance(b['/Top'], NullObject):
+                            if '/Top' in b and not isinstance(b['/Top'], NullObject):
                                 args.append(FloatObject(b['/Top']))
                             else:
                                 args.append(FloatObject(0))
@@ -361,7 +375,7 @@ class PdfFileMerger(object):
             pageno = None
             np = nd['/Page']
             
-            if type(np) == NumberObject:
+            if isinstance(np, NumberObject):
                 continue
             
             for p in pages:
@@ -371,21 +385,21 @@ class PdfFileMerger(object):
             if pageno != None:
                 nd[NameObject('/Page')] = NumberObject(pageno)
             else:
-                raise ValueError, "Unresolved named destination '%s'" % (nd['/Title'],)
+                raise ValueError("Unresolved named destination '%s'" % (nd['/Title'],))
     
     def _associate_bookmarks_to_pages(self, pages, bookmarks=None):
         if bookmarks == None:
             bookmarks = self.bookmarks
 
         for b in bookmarks:
-            if type(b) == list:
+            if isinstance(b, list):
                 self._associate_bookmarks_to_pages(pages, b)
                 continue
                 
             pageno = None
             bp = b['/Page']
             
-            if type(bp) == NumberObject:
+            if isinstance(bp, NumberObject):
                 continue
                 
             for p in pages:
@@ -395,18 +409,18 @@ class PdfFileMerger(object):
             if pageno != None:
                 b[NameObject('/Page')] = NumberObject(pageno)
             else:
-                raise ValueError, "Unresolved bookmark '%s'" % (b['/Title'],)
+                raise ValueError("Unresolved bookmark '%s'" % (b['/Title'],))
                 
     def findBookmark(self, bookmark, root=None):
     	if root == None:
     		root = self.bookmarks
     	
     	for i, b in enumerate(root):
-    		if type(b) == list:
+    		if isinstance(b, list):
     			res = self.findBookmark(bookmark, b)
     			if res:
     				return [i] + res
-    		if b == bookmark or b['/Title'] == bookmark:
+    		elif b == bookmark or b['/Title'] == bookmark:
     			return [i]
     
     	return None
@@ -420,7 +434,7 @@ class PdfFileMerger(object):
 
         if parent == None:
         	iloc = [len(self.bookmarks)-1]
-        elif type(parent) == list:
+        elif isinstance(parent, list):
         	iloc = parent
         else:
         	iloc = self.findBookmark(parent)
@@ -434,10 +448,11 @@ class PdfFileMerger(object):
         	for i in iloc[:-1]:
         		bmparent = bmparent[i]
         	npos = iloc[-1]+1
-        	if npos < len(bmparent) and type(bmparent[npos]) == list:
+        	if npos < len(bmparent) and isinstance(bmparent[npos], list):
         		bmparent[npos].append(dest)
         	else:
         		bmparent.insert(npos, [dest])
+        return dest
         		
         
     def addNamedDestination(self, title, pagenum):
@@ -473,8 +488,8 @@ class OutlinesObject(list):
         bookmark = TreeObject()
 
         bookmark.update({
-            NameObject('/A') : actionRef,
-            NameObject('/Title') : createStringObject(title),
+            NameObject('/A'): actionRef,
+            NameObject('/Title'): createStringObject(title),
         })
 
         pdf._addObject(bookmark)
